# $Id: StochasticSegGenerator.py,v 1.6 2005/11/07 19:44:52 nvf Exp $

#
# Preamble.
#

import logging

from glue import lal
from glue import segments
from onasys import LSCsegFindSegGenerator
from onasys import onasysd
import onasys.SegGenerator

#
# A segment generator for coincident data analysis.  This segment generator
# is a modification of the LSCsegFindSegGenerator that performs N
# LSCdataFind queries, one for each instrument in the list of instruments
# used in the LSCsegFind query.
#
# How to use this.  This segment generator plugin has a [seggenerator]
# section in the .ini file whose contents are identical to the
# [seggenerator] section used by the LSCsegFindSegGenerator plug-in.  If
# LSCdataFind queries are enabled, then rather than a single [datafind]
# section being required, this plug-in requires one such section for each
# instrument.  These sections in the .ini file must have names of the form
# [datafindH], [datafindL], etc., (for Hanford, Livingston, etc.).
# Otherwise, the contents of these [datafind?] sections are identical to
# the contents ofLSCsegFindSegGenerator's [datafind] section
#

#
# A hacked version of onasysd.DataFindConfig that allows the section name
# to be adjusted.
#

class DataFindConfig(onasysd.DataFindConfig):
	def __init__(self, filename, section = "datafind"):
		self.section = section
		onasysd.DataFindConfig.__init__(self, filename)

#
# Here is the SegGeneratorConfig class that Onasys requires this plugin to
# provide.  We subclass LSCsegFindSegGenerator's SegGeneratorConfig class
# in order to get most of the work done for us.  We provide our own parse()
# method, though, in order to do the [datafind] sections our own way.
#

class SegGeneratorConfig(LSCsegFindSegGenerator.SegGeneratorConfig):
	def parse(self):
		onasysd.SegFindConfig.parse(self)
		self.initial_epoch = lal.LIGOTimeGPS(self.parser.get(self.section, "initial_epoch"))
		self.datafind = []
                instruments = self.instrument.split(",")
                instruments.sort() # sort returns nothing; sorts in-place
		for inst in instruments: 
			try:
				# skip ones we've already done
				if datafindconfig.observatory == inst[0]:
					continue
			except:
				pass
			datafindconfig = DataFindConfig(self.filename, section = "datafind" + inst[0])
			datafindconfig.parse()
			self.datafind.append(onasysd.DataFind(datafindconfig))

	def log(self):
		l = onasys.SegGenerator.SegGeneratorConfig.log(self)
		for datafind in self.datafind:
			l += datafind.config.log()
		return l

	def validate(self):
		for datafind in self.datafind:
			datafind.config.validate()


#
# Finally we provide the SegGenerator class that Onasys requires this
# plugin to provide.  Again, we base it off of
# LSCsegFindSegGenerator.SegGenerator in order to get most of the work done
# for us.  The only class method we have to change is generate(), and it's
# mostly the same as the one in LSCsegFindSegGenerator.SegGenerator.
#

class SegGenerator(LSCsegFindSegGenerator.SegGenerator):
	def generate(self, available, clock):
		# Determine the window in which to search for segments.
		self.search_window.next(self.config.initial_epoch, clock.maxend(), self.config.min_query_length)

		# Put a useful message in the log
		logging.info(self.segfind.command(self.search_window.window))

		# Do an LSCsegFind query
		try:
			list = self.segfind.list(self.search_window.window)
		except Exception, e:
			raise onasys.SegGenerator.SegGeneratorError, "LSCsegFind query failure: " + str(e)
		logging.info("LSCsegFind response = " + str(list))

		# Do the LSCdataFind queries.
		try:
			for datafind in self.config.datafind:
				# From (what's left of) our list, determine
				# the segments we haven't seen before
				new = list - available.get("available", segments.segmentlist())

				# For each of these, check for frame files.
				datafindlist = segments.segmentlist()
				for seg in new:
					logging.info(datafind.command(seg))
					datafindlist |= datafind.list(seg)

				# Trim the list of coincident segments
				list &= datafindlist
		except Exception, e:
			raise onasys.SegGenerator.SegGeneratorError, "LSCdataFind query failure: " + str(e)

		return {"available": list}
