import xml.parsers.expat
import re
import string


##################################################################
#
#  Condor Log Events
#

class CondorEvent(object):
    SUBMIT                 = 0  # Job submitted  
    EXECUTE                = 1  # Job now running  
    EXECUTABLE_ERROR       = 2  # Error in executable  
    CHECKPOINTED           = 3  # Job was checkpointed  
    JOB_EVICTED            = 4  # Job evicted from machine  
    JOB_TERMINATED         = 5  # Job terminated  
    IMAGE_SIZE             = 6  # Image size of job updated  
    SHADOW_EXCEPTION       = 7  # Shadow threw an exception  
    GENERIC                = 8  # Generic Log Event  
    JOB_ABORTED            = 9  # Job Aborted  
    JOB_SUSPENDED          = 10 # Job was suspended  
    JOB_UNSUSPENDED        = 11 # Job was unsuspended  
    JOB_HELD               = 12 # Job was held  
    JOB_RELEASED           = 13 # Job was released  
    NODE_EXECUTE           = 14 # Parallel Node executed  
    NODE_TERMINATED        = 15 # Parallel Node terminated  
    POST_SCRIPT_TERMINATED = 16 # POST script terminated  
    GLOBUS_SUBMIT          = 17 # Job Submitted to Globus  
    GLOBUS_SUBMIT_FAILED   = 18 # Globus Submit failed  
    GLOBUS_RESOURCE_UP     = 19 # Globus Resource Up  
    GLOBUS_RESOURCE_DOWN   = 20 # Globus Resource Down  
    REMOTE_ERROR           = 21 # Remote Error  

    eventNames = {
            SUBMIT                 : "Job submitted",
            EXECUTE                : "Job now running",
            EXECUTABLE_ERROR       : "Error in executable",
            CHECKPOINTED           : "Job was checkpointed",
            JOB_EVICTED            : "Job evicted from machine",
            JOB_TERMINATED         : "Job terminated",
            IMAGE_SIZE             : "Image size of job updated",
            SHADOW_EXCEPTION       : "Shadow threw an exception",
            GENERIC                : "Generic Log Event",
            JOB_ABORTED            : "Job Aborted",
            JOB_SUSPENDED          : "Job was suspended",
            JOB_UNSUSPENDED        : "Job was unsuspended",
            JOB_HELD               : "Job was held",
            JOB_RELEASED           : "Job was released",
            NODE_EXECUTE           : "Parallel Node executed",
            NODE_TERMINATED        : "Parallel Node terminated",
            POST_SCRIPT_TERMINATED : "POST script terminated",
            GLOBUS_SUBMIT          : "Job Submitted to Globus",
            GLOBUS_SUBMIT_FAILED   : "Globus Submit failed",
            GLOBUS_RESOURCE_UP     : "Globus Resource Up",
            GLOBUS_RESOURCE_DOWN   : "Globus Resource Down",
            REMOTE_ERROR           : "Remote Error"
     }

    def __init__(self, eventtype, cluster, proc, subproc, datetime, message, notes):
        self.eventtype = int(eventtype)
        self.cluster = int(cluster)
        self.proc = int(proc)
        self.subproc = int(subproc)
        self.datetime = datetime
        self.message = message
        self.notes = notes
        self.dagid = None

        if self.eventtype == CondorEvent.SUBMIT:
            # Look for host
            # Look for a dag id
            self.host = self._findHost(message)
            m = re.compile(r'.*DAG Node: ([^\n]+).*',re.MULTILINE).match(notes)
            if m:
                self.dagid = m.group(1)
            else:
                # trying to indicate that this is the main
                # condor_dagman node.
                # hacky!
                self.dagid = "_MAIN_"

        elif self.eventtype == CondorEvent.EXECUTE:
            self.host = self._findHost(message)

        elif self.eventtype == CondorEvent.JOB_TERMINATED:
            m = re.compile(r'.*\(return value (\d+)\).*').match(notes)
            if m:
                self.normal = 1
                self.returnValue = int(m.group(1))
                self.signalNumber = 0
            else:
                # XXX What does this kind of record look like?
                self.normal = 0
                self.returnValue = 0
                self.signalNumber = 0

    def _findHost(self, s):
        m = re.compile(r'.*<(.+):\d+>.*').match(s)
        if m: return m.group(1)
        else: return "UNKNOWN"

    def type(self):
        return self.eventtype

    def typeName(self):
        try:
            name = CondorEvent.eventNames[self.type()]
        except:
            return "UNKNOWN CONDOR EVENT TYPE '%s'" % str(self.type())
        return name

    def __str__(self):
        return "EVENT: %s %d/%s" % (self.typeName(), self.cluster, self.dagid)



##################################################################
#
#  Condor Log Parsing
#

# Standard Condor logs look like:
#
# %3d (%d.%d.%d) %d/%d %d:%d:%d %s
#  |    |  |  |    |       |     |
#  |    |  |  |    |       |     +- Message
#  |    |  |  |    |       +------- Time
#  |    |  |  |    +--------------- Date
#  |    |  |  +-------------------- Sub-Process
#  |    |  +----------------------- Process
#  |    +-------------------------- Cluster
#  +------------------------------- Event Code
#
# Followed by lines of text terminated by a line containing only:
#
# ...
#

#  XML Condor Logs...
#
#  e.g.
#
# <c>
#     <a n="MyType"><s>SubmitEvent</s></a>
#     <a n="EventTypeNumber"><i>0</i></a>
#     <a n="EventTime"><s>2004-10-26T13:36:44</s></a>
#     <a n="Cluster"><i>710235</i></a>
#     <a n="Proc"><i>0</i></a>
#     <a n="Subproc"><i>0</i></a>
#     <a n="SubmitHost"><s>&lt;129.89.201.233:38917&gt;</s></a>
#     <a n="LogNotes"><s>DAG Node: A</s></a>
# </c>
# <c>
#     <a n="MyType"><s>ExecuteEvent</s></a>
#     <a n="EventTypeNumber"><i>1</i></a>
#     <a n="EventTime"><s>2004-10-26T13:36:45</s></a>
#     <a n="Cluster"><i>710235</i></a>
#     <a n="Proc"><i>0</i></a>
#     <a n="Subproc"><i>0</i></a>
#     <a n="ExecuteHost"><s>&lt;129.89.201.233:38917&gt;</s></a>
# </c>
# <c>
#     <a n="MyType"><s>JobTerminatedEvent</s></a>
#     <a n="EventTypeNumber"><i>5</i></a>
#     <a n="EventTime"><s>2004-10-26T13:36:45</s></a>
#     <a n="Cluster"><i>710235</i></a>
#     <a n="Proc"><i>0</i></a>
#     <a n="Subproc"><i>0</i></a>
#     <a n="TerminatedNormally"><b v="t"/></a>
#     <a n="ReturnValue"><i>0</i></a>
#     <a n="RunLocalUsage"><s>Usr 0 00:00:00, Sys 0 00:00:00</s></a>
#     <a n="RunRemoteUsage"><s>Usr 0 00:00:00, Sys 0 00:00:00</s></a>
#     <a n="TotalLocalUsage"><s>Usr 0 00:00:00, Sys 0 00:00:00</s></a>
#     <a n="TotalRemoteUsage"><s>Usr 0 00:00:00, Sys 0 00:00:00</s></a>
#     <a n="SentBytes"><r>0.000000000000000E+00</r></a>
#     <a n="ReceivedBytes"><r>0.000000000000000E+00</r></a>
#     <a n="TotalSentBytes"><r>0.000000000000000E+00</r></a>
#     <a n="TotalReceivedBytes"><r>0.000000000000000E+00</r></a>
# </c>



class CondorLogParser(object):

    STD_PAT = re.compile(r'(?P<id>\d\d\d) \((?P<cluster>\d+)\.(?P<process>\d+)\.(?P<subprocess>\d+)\) (?P<date>[^ ]+) (?P<time>[^ ]+) (?P<message>.*)')

    DELIM_PATTERN = re.compile(r'(?P<text>.*?)(?P<delim>\n...\n|\n</c>\n)(?P<rest>.*)',re.MULTILINE|re.DOTALL)
    STD_DELIM = '\n...\n'
    XML_DELIM = '\n</c>\n'

    def __init__(self, eventHandler):
        self.eventHandler = eventHandler
        self.buf = ""
        # Set up an XML parser in case we get XML log data.
        self.xmlParser = xml.parsers.expat.ParserCreate()
        self.xmlParser.StartElementHandler  = self.start_element
        self.xmlParser.EndElementHandler    = self.end_element
        self.xmlParser.CharacterDataHandler = self.char_data
        try:
            self.xmlParser.buffer_text = 1
        except: pass
        self.xmlParser.Parse("<log>") # prime the parser

    def start_element(self, name, attrs):
        if name == 'c':
            self.xmlRec = {}
        elif name == 'a':
            self.xmlField = str(attrs['n'])

    def end_element(self, name):
        if name == 'i':
            self.xmlRec[self.xmlField] = int(self.xmlData)
        elif name == 's':
            self.xmlRec[self.xmlField] = str(self.xmlData)
        elif name == 'r':
            self.xmlRec[self.xmlField] = float(self.xmlData)
        elif name == 'c':
            #eventtype, cluster, proc, subproc, date, time, message, notes
            message = '<123.32.12.3:323>'
            try:
                notes = self.xmlRec['LogNotes']
            except:
                notes = ""
            event = CondorEvent(
                self.xmlRec['EventTypeNumber'],
                self.xmlRec['Cluster'],
                self.xmlRec['Proc'],
                self.xmlRec['Subproc'],
                self.xmlRec['EventTime'],
                message,
                notes)
            self.eventHandler(event)

    def char_data(self, theData):
        self.xmlData = theData

    def parseStdRecord(self, recordData):
        lines = recordData.split('\n')
        m = self.STD_PAT.match(lines[0])
        if m:
            eventId = int(m.group("id"))
            cluster = int(m.group("cluster"))
            process = int(m.group("process"))
            subprocess = int(m.group("subprocess"))
            date = m.group("date")
            time = m.group("time")
            message = m.group("message")
            notes = string.join(lines[1:],"\n")
            event= CondorEvent(eventId, cluster, process, subprocess, \
                               date+" "+time, message, notes)
            self.eventHandler(event)
        else:
            # This is seriously bad.
            print "BADNESS", lines[0]
            return None

    def parse(self, data):
        allData = self.buf + data
        while allData:
            match = self.DELIM_PATTERN.match(allData)
            if not match:
                #print "UGH", allData
                self.buf = allData
                return
            delim = match.group('delim')
            if delim == self.XML_DELIM:
                self.xmlParser.Parse(match.group('text'))
                self.xmlParser.Parse(delim)
            elif delim == self.STD_DELIM:
                self.parseStdRecord(match.group('text'))
            else:
                raise Exception("What delimiter is this?! (%s)"%delim)
            allData = match.group('rest')
        self.buf = allData

